import aiohttp
from aiohttp import hdrs
from aiohttp.helpers import AccessLogger
from aiohttp.web_server import RequestHandler

from aiokts.error import ServerError
from aiokts.response import ApiErrorResponse
from aiokts.util.json_utils import json_dumps


class KtsAccessLogger(AccessLogger):
    def log_request(self, request, message, environ, response, transport, time):
        """Log access.

        :param message: Request object. May be None.
        :param environ: Environment dict. May be None.
        :param response: Response object.
        :param transport: Tansport object. May be None
        :param float time: Time taken to serve the request.
        """
        
        ctx = request.ctx if hasattr(request, 'ctx') else None
        if ctx is None:
            return self.log(message, environ, response, transport, time)
        
        try:
            fmt_info = self._format_line(
                [message, environ, response, transport, time])
            
            values = list()
            extra = dict()
            for key, value in fmt_info:
                values.append(value)
                
                if key.__class__ is str:
                    extra[key] = value
                else:
                    extra[key[0]] = {key[1]: value}
            
            ctx.logger.info(self._log_format % tuple(values), extra=extra)
        except Exception:
            ctx.logger.exception("Error in logging")


class KtsRequestHandler(RequestHandler):
    def __init__(self, *args, **kwargs):
        access_log = kwargs.get('access_log')
        access_log_format = kwargs.get('access_log_format')
        super(KtsRequestHandler, self).__init__(*args, **kwargs)
        
        if access_log:
            self.access_logger = KtsAccessLogger(access_log, access_log_format)
        else:
            self.access_logger = None
        self._middlewares = None
    
    def log_access(self, message, environ, response, time):
        if self.access_logger:
            if self._request is not None:
                self.access_logger.log_request(self._request, message, environ, response, self.transport, time)
            else:
                self.access_logger.log(message, environ, response, self.transport, time)
    
    def handle_error(self, status=500, message=None,
                     payload=None, exc=None, headers=None, reason=None):
        """rewrite of super() function

        Handle errors.

        Returns HTTP response with specific status code. Logs additional
        information. It always closes current connection."""
        now = self._loop.time()
        try:
            if self.transport is None:
                # client has been disconnected during writing.
                return ()
            
            self.log_exception("Error handling request")
            
            error = ServerError.INTERNAL_ERROR
            data = ApiErrorResponse.generate_response_dict(
                message=error.message,
                data=error.payload
            )
            data = json_dumps(data).encode()
            
            response = aiohttp.Response(self.writer, status, close=True)
            response.add_header(hdrs.CONTENT_TYPE, 'application/json; charset=utf-8')
            response.add_header(hdrs.CONTENT_LENGTH, str(len(data)))
            if headers is not None:
                for name, value in headers:
                    response.add_header(name, value)
            response.send_headers()
            
            response.write(data)
            # disable CORK, enable NODELAY if needed
            self.writer.set_tcp_nodelay(True)
            drain = response.write_eof()
            
            self.log_access(message, None, response, self._loop.time() - now)
            return drain
        finally:
            self.keep_alive(False)



import logging
import uuid
import weakref

from aiohttp.helpers import reify

__all__ = [
    'ContextDataObject',
    'Context'
]


class ContextLogger(logging.Logger):
    def __init__(self, ctx, name='ctxLogger', parent=logging.root):
        super(ContextLogger, self).__init__(name)
        self._ctx = ctx
        self.parent = parent
        
    @property
    def ctx(self):
        return self._ctx
        
    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False):
        msg = '{}{}{}'.format(self.ctx.log_prepend, msg, self.ctx.log_append)
        return super(ContextLogger, self)._log(level, msg, args, exc_info, extra, stack_info)


class ContextDataObject(object):
    def __repr__(self):
        return '<ContextDataObject keys_count:{}>'.format(len(self.__dict__))
    
    def clear(self):
        self.__dict__.clear()
        
    def __getattr__(self, item):
        return None
    

class Context(object):
    CONTEXT_DATA_OBJECT_CLS = ContextDataObject
    
    __slots__ = [
        '_request',
        '_view',
        'hash',
        'logger',
        '_data',
        '_cache'
    ]
    
    def __init__(self, request):
        self._request = weakref.ref(request) if request is not None else None
        self.hash = self._generate_hash()
        self.logger = ContextLogger(self)
        self._data = self.CONTEXT_DATA_OBJECT_CLS()

        self._cache = {}
        
    @property
    def request(self):
        return self._request() if self._request is not None else None
    
    def log_request(self):
        if self._request is not None:
            self.logger.info('Request {method} {path}{query}'.format(
                method=self.request.method,
                path=self.request.path,
                query=self.request.query_string if not self.request.query_string else '?' + self.request.query_string
            ))
    
    @staticmethod
    def _generate_hash():
        h = uuid.uuid4().hex
        h = h[:10]
        return h
    
    @property
    def data(self):
        return self._data
    
    @reify
    def log_prepend(self):
        """
            This property can return a string that will be prepended to a log string (to the start)
            
            In base Context it has @reify decorator from aiohttp to cache the result indefinitely,
            but if you subclass Context you are allowed to use whatever @property decorator you want
        """
        return '[ctx:{}] '.format(self.hash)

    @reify
    def log_append(self):
        """
            This property can return a string that will be appended to a log string (to the end)

            In base Context it has @reify decorator from aiohttp to cache the result indefinitely,
            but if you subclass Context you are allowed to use whatever @property decorator you want
        """
        return ''



import abc
import json

import datetime
import time

_features = {
    'bson_object_id': False
}

try:
    from bson import ObjectId
    _features['bson_object_id'] = True
except ImportError:
    pass


class CustomJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, JsonSerializable):
            return obj.__to_json__()
        
        if isinstance(obj, datetime.datetime) or isinstance(obj, datetime.date):
            return int(time.mktime(obj.timetuple()))
        
        if _features['bson_object_id']:
            if isinstance(obj, ObjectId):
                return str(obj)
        
        return super(CustomJsonEncoder, self).default(obj)


class JsonSerializable:
    @abc.abstractmethod
    def __to_json__(self):
        raise NotImplementedError()


def json_dumps(obj, compact=True, **kwargs):
    kwargs['cls'] = kwargs.get('cls', CustomJsonEncoder)
    kwargs['ensure_ascii'] = kwargs.get('ensure_ascii', False)
    if compact:
        kwargs['separators'] = (',', ':')
    return json.dumps(obj, **kwargs)


def json_dump(fp, obj, compact=True, **kwargs):
    kwargs['cls'] = kwargs.get('cls', CustomJsonEncoder)
    kwargs['ensure_ascii'] = kwargs.get('ensure_ascii', False)
    if compact:
        kwargs['separators'] = (',', ':')
    return json.dump(fp, obj, **kwargs)


def json_loads(obj, **kwargs):
    kwargs['encoding'] = kwargs.get('encoding', 'utf-8')
    return json.loads(obj, **kwargs)


def json_load(fp, **kwargs):
    kwargs['encoding'] = kwargs.get('encoding', 'utf-8')
    return json.load(fp, **kwargs)


def ensure_json(obj, **kwargs):
    if obj is None:
        return None
    if isinstance(obj, list) or isinstance(obj, dict):
        return obj
    return json_loads(obj, **kwargs)



import copy


class Error:
    def __init__(self, code, message: str, payload: dict = None, http_code=None):
        self.code = code
        self.message = message
        self.payload = payload
        self.http_code = http_code
        self.extra = None
    
    def __call__(self, message=None, payload=None, extra=None):
        c = copy.deepcopy(self)
        if message is not None:
            c.message = message
        if payload is not None:
            c.payload = payload
        if extra is not None:
            c.extra = extra
        return c
    
    def __repr__(self):
        return '<{} (code={}, message={}, extra={})>'.format(self.__class__.__name__,
                                                             self.code,
                                                             self.message,
                                                             self.extra)
    
    def __str__(self):
        return self.message


class ServerError(Exception):
    NOT_AUTHORIZED = Error('not_authorized', 'Not Authorized', http_code=401)
    BAD_REQUEST = Error('bad_request', 'Bad Request', http_code=400)
    FORBIDDEN = Error('forbidden', 'Forbidden', http_code=403)
    NOT_FOUND = Error('not_found', 'Not Found', http_code=404)
    INTERNAL_ERROR = Error('internal_error', 'Internal Server Error', http_code=500)
    
    def __init__(self, error: Error):
        self.error = error
    
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, repr(self.error))
    
    def __str__(self):
        return str(self.error)



import asyncio
import json

from aiohttp import web
from aiohttp.web_exceptions import HTTPNotFound, HTTPException, HTTPMethodNotAllowed

from aiokts.context import Context
from aiokts.error import ServerError
from aiokts.response import ApiErrorResponse
from aiokts.util.arguments import ArgumentException


class BaseView(web.View):
    CONTEXT_CLS = Context

    def __init__(self, request):
        super().__init__(request)
        self.app = self.request.app
        self.ctx = self.request.ctx
        self.request_data = None

    @property
    def logger(self):
        return self.ctx.logger if self.ctx is not None else None

    @property
    def stores(self):
        return self.app.stores

    @property
    def loop(self):
        return self.app.loop

    @asyncio.coroutine
    def __iter__(self):
        try:
            yield from self._parse_request()
            yield from self.pre_handle()
            return (yield from super(BaseView, self).__iter__())
        except Exception as e:
            res = self.handle_exception(e)
            return res

    async def _parse_request(self):
        if self.request.method == 'GET':
            source = self.request.url.query
        else:
            if self.request.content_type.startswith('application/json'):
                try:
                    source = await self.request.json()
                except json.JSONDecodeError:
                    raise ServerError(ServerError.BAD_REQUEST(message='Body must be a valid json'))
            else:
                source = await self.request.post()
        self.request_data = source

    async def pre_handle(self):
        pass

    def handle_exception(self, e):
        if isinstance(e, ServerError):
            error = e.error
            if error.http_code is None:
                status = 500
            else:
                status = error.http_code
            self.logger.error('[{}] Error: {}'.format(repr(error.code), error.message))

            return ApiErrorResponse(message=error.message,
                                    data=error.payload,
                                    http_status=status,
                                    code=error.code)
        elif isinstance(e, ArgumentException):
            self.logger.warning('ArgumentException: {}'.format(e.message))
            return ApiErrorResponse(message='Field {} is invalid'.format(e.field) if e.field is not None else '',
                                    system_message=e.message,
                                    http_status=400,
                                    code='bad_request',
                                    data=dict(field=e.field))
        elif isinstance(e, HTTPException):
            if e.status_code >= 500:
                self.logger.exception('HTTPException: {}'.format(str(e)), exc_info=e)
            e.headers.pop('Content-Type')
            e.headers.pop('Content-Length')
            return ApiErrorResponse(message=e.reason,
                                    http_status=e.status_code,
                                    headers=e.headers)
        elif isinstance(e, Exception):
            self.logger.exception('Exception: {}'.format(str(e)), exc_info=e)
            return ApiErrorResponse(message='Internal Server Error')
        else:
            self.logger.error('{} is not an exception'.format(repr(e)))


class ActionBaseView(BaseView):
    CONTEXT_CLS = Context

    def __init__(self, request):
        super().__init__(request)

    @property
    def store(self):
        return self.app.store

    @property
    def default_methods(self):
        """
            Dict with keys for HTTP Methods (GET, POST, ...)
            and values are dict with mapping action => func
            i.e.
            {
                "GET": self.default_get,
                "POST": self.default_post
            }
        """
        return {}

    @property
    def methods(self):
        """
            Dict with keys for HTTP Methods (GET, POST, ...)
            and values are dict with mapping action => func
            i.e.
            {
                "GET": {
                    "home": self.home
                },
                "POST": {
                    "post": self.post
                }
            }
        """
        return {}

    async def before_action(self):
        pass

    async def after_action(self):
        pass

    @asyncio.coroutine
    def __iter__(self):
        try:
            yield from self._parse_request()
            yield from self.pre_handle()

            action_title = self.request.match_info['method']

            if action_title is None:
                raise HTTPNotFound()

            method = self.request.method.upper()
            methods = self.methods
            if methods is None:
                methods = {}

            actions = methods.get(method, {})

            def_method = self.default_methods.get(method)
            if action_title in actions:
                executing_method = actions.get(action_title)
            elif def_method is not None:
                executing_method = def_method
            else:
                executing_method = None

            if executing_method is not None:
                yield from self.before_action()
                result = yield from executing_method()
                yield from self.after_action()
                return result
            else:
                allowed_methods = []
                for k, k_actions in methods.items():
                    if action_title in k_actions:
                        allowed_methods.append(k)

                if allowed_methods:
                    raise HTTPMethodNotAllowed(method, allowed_methods=allowed_methods)

                raise HTTPNotFound()
        except Exception as e:
            res = self.handle_exception(e)
            return res



import abc
import asyncio
import logging
import logging.config
import traceback

from aiokts.storeset import StoreSet


class Daemon(object):
    def __init__(self, loop=None, **kwargs):
        self.name = self.__class__.__name__
        self.loop = loop
        self._stopped = False
        self._run_task = None
        self.stores = StoreSet(kwargs.get('stores', {}))

    @property
    def is_stopped(self):
        return self._stopped

    @property
    def is_running(self):
        return not self._stopped

    @asyncio.coroutine
    def start(self):
        def on_run_finished(future):
            if not future.cancelled():
                exc = future.exception()
                if exc:
                    future.print_stack()
                    logging.error('Daemon {} exception'.format(self.__class__.__name__))
                self._stopped = True
            else:
                logging.info('Daemon {} cancelled'.format(self.__class__.__name__))
                asyncio.ensure_future(self.stop(), loop=self.loop)

        try:
            self._run_task = asyncio.ensure_future(self.run(), loop=self.loop)
            yield from self._run_task
        except asyncio.futures.CancelledError:
            pass
        except Exception as e:
            logging.exception(e)

    @asyncio.coroutine
    def stop(self):
        logging.info('Stopping {} daemon'.format(self.name))
        if asyncio.iscoroutine(self.cancel) or asyncio.iscoroutinefunction(self.cancel):
            yield from self.cancel()
        else:
            self.cancel()
        if self._run_task is not None:
            self._run_task.cancel()
            self._run_task = None
        self._stopped = True
        logging.info('Stopped {} daemon'.format(self.name))

    @abc.abstractmethod
    @asyncio.coroutine
    def run(self):
        raise NotImplementedError()

    def cancel(self):
        pass


def main(daemon):
    loop = asyncio.get_event_loop()

    try:
        loop.run_until_complete(daemon.start())
        loop.run_until_complete(daemon.stop())
        return 0
    except KeyboardInterrupt:
        loop.run_until_complete(daemon.stop())
        return 0
    except:
        return 1



from aiohttp.web import Response

from aiokts.util.json_utils import json_dumps

__all__ = [
    'JsonResponse',
    'ApiResponse',
    'ApiOkResponse',
    'ApiErrorResponse'
]


class JsonResponse(Response):
    def __init__(self, body=None, *,
                 status=200,
                 reason=None,
                 text=None,
                 headers=None,
                 charset=None,
                 json_dump_func=None):
        if json_dump_func is None:
            json_dump_func = json_dumps
            
        body, status = self._process_body(body, status)
        
        if charset is None:
            charset = 'utf-8'
        
        body = json_dump_func(body).encode(charset)
        
        super().__init__(
            body=body,
            status=status,
            reason=reason,
            text=text,
            headers=headers,
            content_type='application/json',
            charset=charset
        )
    
    def _process_body(self, body, status=200):
        return body, status


class RawApiResponse(JsonResponse):
    pass


class ApiResponse(RawApiResponse):
    def __init__(self, status='ok', data=None, http_status=200, *,
                 headers=None, charset=None, json_dump_func=None, **kwargs):
        """
        :param status: "ok" or "error"
        :param data: data payload
        :param http_status: http code
        :param kwargs: extra kwargs will be inlined in the root response, so
                      if extra = { 'code': 'hello' } then response will be
                      {
                        'status': status,
                        'code': 'hello',
                        'data': data
                      }
        """
        if data is None:
            data = {}
        
        self._api_status = status
        self._api_data = data
        self._api_http_status = http_status
        self._api_extra = kwargs

        body = self.__class__.generate_response_dict(
            data=self._api_data,
            **self._api_extra
        )
        
        super(ApiResponse, self).__init__(
            body=body,
            status=self._api_http_status,
            headers=headers,
            charset=charset,
            json_dump_func=json_dump_func
        )
        
    @staticmethod
    def generate_response_dict(api_status='ok', data=None, **extra):
        if data is None:
            data = {}
        res = {
            'status': api_status,
            'data': data
        }
        res.update(extra)
        return res


class ApiOkResponse(ApiResponse):
    def __init__(self, data=None, http_status=200, *,
                 headers=None, charset=None, json_dump_func=None, **kwargs):
        super(ApiOkResponse, self).__init__('ok', data, http_status,
                                            headers=headers,
                                            charset=charset,
                                            json_dump_func=json_dump_func, **kwargs)

    @staticmethod
    def generate_response_dict(data=None, **extra):
        return ApiResponse.generate_response_dict(api_status='ok', data=data, **extra)


class ApiErrorResponse(ApiResponse):
    def __init__(self, message=None, data=None, http_status=500, *,
                 headers=None, charset=None, json_dump_func=None, **kwargs):
        if message:
            kwargs['message'] = message
        super(ApiErrorResponse, self).__init__('error', data, http_status,
                                               headers=headers,
                                               charset=charset,
                                               json_dump_func=json_dump_func, **kwargs)
    
    @staticmethod
    def generate_response_dict(message=None, data=None, **extra):
        if message:
            extra['message'] = message
        return ApiResponse.generate_response_dict(api_status='error', data=data, **extra)



class BaseStore(object):
    __slots__ = ["_obj", "__weakref__"]

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, "_obj"), name)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_obj"), name, value)

    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_obj"))

    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_obj"))

    #
    # factories
    #
    _special_names = [
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__',
        '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__',
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__',
        '__getslice__', '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__',
        '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__',
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__',
        '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__',
        '__neg__', '__oct__', '__or__', '__pos__', '__pow__', '__radd__',
        '__rand__', '__rdiv__', '__rdivmod__', '__reduce__', '__reduce_ex__',
        '__repr__', '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__',
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__',
        '__rtruediv__', '__rxor__', '__setitem__', '__setslice__', '__sub__',
        '__truediv__', '__xor__', 'next',
    ]

    @classmethod
    def _create_class_proxy(cls, theclass):
        """creates a proxy for the given class"""

        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, "_obj"), name)(*args, **kw)

            return method

        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name):
                namespace[name] = make_method(name)
        return type("%s(%s)" % (cls.__name__, theclass.__name__), (cls,), namespace)

    def __new__(cls, obj, *args, **kwargs):
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(obj.__class__)
        ins = object.__new__(theclass)
        theclass.__init__(ins, obj, *args, **kwargs)
        return ins



import functools

_sentinel = object()


class Argument:
    def __init__(self, required=False, default=None, type=None, validator=None, validator_message='', filter=None):
        self.required = required
        self.default = default
        self.type = type
        self.validator = validator
        self.validator_message = validator_message
        self.filter = filter


class ArgumentException(Exception):
    def __init__(self, field, message=_sentinel):
        super(ArgumentException, self).__init__(message)
        self.field = field
        if message is _sentinel:
            if self.field is not None:
                self.message = 'Argument `{}` is required'.format(self.field)
            else:
                self.message = 'Unknown problem with arguments'
        else:
            self.message = message


def arguments(arglist):
    """:Валидатор/фильтратор входных параметров

    Проверяет переданные параметры и отфильтровывает все лишние.
    Аргумент считается не указанным либо если его просто нет в входных аргментах, либо если он None

    Пример использования:
    <pre>
    @arguments({
        'arg_name': Argument(required=False, default=7, type=int, validator=lambda x: x > 5, validator_message='must be greater than 5'),
        ...
    })
    def func(self, *args, **kwargs):
        ...

    func(arg_name=1, ...)
    </pre>

    - required (bool)
        Требовать наличия параметра (кидать эксепшн) или тихо игнорировать его отсутствие

    - default
        Дефолтное значение в случае отсутствия параметра и required=False

    - type
        Требуемый тип значения, или None если проверять не надо

    - validator
        функция-валидатор, которая применяется к значению. Если возвращает не True, райзится эксепшн.
        Применяется также и к дефолтному значению. None, если не нужен никакой этот ваш валидатор

    - validator_message
        сообщение, которое выводится в эксепшне, если значение аргумента не проходит проверку валидатором

    - filter
        функция-фильтр, которая принимает на вход переданное значение аргумента и возвращает то, которое будет
        передано в функцию. Например, так можно преобразовывать флаги из dict в int

    :raises ArgumentException
    """

    def _arguments(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            checked_kwargs = check_arguments(arglist, kwargs)
            return func(*args, **checked_kwargs)

        inner._has_arguments_ = True
        inner.arglist = arglist
        return inner

    return _arguments


def check_arguments(arglist, kwargs, *, cast_type=False):
    if not isinstance(arglist, dict):
        raise ArgumentException(None, '@arguments expects arg definition as a dict with `Argument` class values')

    filtered_kwargs = {}
    for arg_name, arg_definition in arglist.items():
        if not isinstance(arg_definition, Argument):
            raise ArgumentException(None, '@arguments expects arg definition as a dict with `Argument` class values')

        # check argument existence
        is_default_value = False
        if arg_name not in kwargs and arg_definition.required:
            raise ArgumentException(arg_name, '`%s` argument is required' % arg_name)
        else:
            # not required argument is not specified
            arg_value = kwargs.get(arg_name)
            if arg_value is None:
                if arg_definition.default is not None:
                    arg_value = arg_definition.default
                    is_default_value = True
                else:
                    arg_value = None

        # check argument type
        if arg_definition.type is not None and (
            arg_definition.required and arg_value is not None):  # None means "any type, do not check"
            if not cast_type:
                if not isinstance(arg_value, arg_definition.type):
                    raise ArgumentException(arg_name,
                                            '`%s` must be `%s`, but got `%s`' %
                                            (arg_name, str(arg_definition.type), str(type(arg_value))))
            else:
                try:
                    arg_value = arg_definition.type(arg_value)
                except Exception as e:
                    raise ArgumentException(arg_name,
                                            'Casting `%s` to type `%s` (which has type `%s`) failed: `%s`' %
                                            (arg_name, str(arg_definition.type), str(type(arg_value)), str(e))) from e

        # filter value
        if arg_value is not None and callable(arg_definition.filter):
            arg_value = arg_definition.filter(arg_value)

        # validate value
        if arg_value is not None and callable(arg_definition.validator):
            if not arg_definition.validator(arg_value):
                raise ArgumentException(arg_name,
                                        '%svalue `%s` for argument `%s` was rejected by validator: `%s`' %
                                        ('default ' if is_default_value else '', str(arg_value),
                                         arg_name,
                                         str(arg_definition.validator_message)))

        # all checks passed
        filtered_kwargs[arg_name] = arg_value
    return filtered_kwargs


def has_arguments(func):
    return func.__dict__.get('_has_arguments_', False) and hasattr(func, 'arglist')



import aiotarantool
from datetime import datetime

from aiokts.stores.base import BaseStore
import logging

logger = logging.getLogger('tarantool')


class TarantoolStore(BaseStore):
    def __init__(self, config):
        logger.debug(config)
        tnt = aiotarantool.Connection(
            config['host'],
            config['port'],
            user=config.get('user'), password=config.get('password'),
            connect_now=True
        )
        super().__init__(tnt)

    def call(self, name, *args):
        start = datetime.now()
        import os
        logger.debug('Tarantool call begin: {}, pid: {}, user: {}'.format(name, os.getpid(), self.user))

        result = super().__getattribute__('call')(name, *args)
        end = datetime.now()
        delta = end - start
        delta = delta.seconds + (delta.microseconds / (10 ** 6))
        logger.debug('Tarantool call end: {}, time: {}'.format(name, delta))

        return result

    def __getattribute__(self, name):
        if name != 'call':
            return super().__getattribute__(name)
        else:
            def wrap(name, *args):
                return TarantoolStore.call(self, name, *args)

            return wrap



from aiohttp.web_server import Server

from aiokts.handler import KtsRequestHandler


class KtsServer(Server):
    def __call__(self):
        return KtsRequestHandler(
            self, loop=self._loop,
            **self._kwargs)



from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
version = '0.2.1'
modules = [
    'aiohttp>=1.2.0'
]

setup(
    name='aiokts',
    version=version,
    description='Tuned asyncio and aiohttp classes for simpler creation of powerful APIs',
    long_description='Tuned asyncio and aiohttp classes for simpler creation of powerful APIs',

    author='KTS',
    author_email='aiokts@ktsstudio.ru',
    url='https://github.com/KTSStudio/aiokts',
    download_url='https://github.com/KTSStudio/aiokts/tarball/v' + version,
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
    ],

    keywords=' '.join(['aiohttp', 'asyncio', 'python 3.5', 'api']),
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=modules,
)



import warnings
import weakref

from aiohttp import web
from aiohttp.log import web_logger

from aiokts.context import Context
from aiokts.request import KtsRequest
from aiokts.server import KtsServer
from aiokts.storeset import StoreSet

sentinel = object()


class KtsHttpApplication(web.Application):
    ROUTES = []
    SERVER_CLS = KtsServer
    
    def __init__(self, **kwargs):
        super().__init__(
            logger=kwargs.get('logger', web_logger),
            loop=kwargs.get('loop'),
            router=kwargs.get('router'),
            middlewares=kwargs.get('middlewares', ()),
            debug=kwargs.get('debug', False)
        )
        self.stores = StoreSet(kwargs.get('stores', {}))
        
        for route in self.ROUTES:
            method, path, view_cls = route
            self.router.add_route(method, path, view_cls)

    def _make_request(self, message, payload, protocol,
                      _cls=KtsRequest):
        req = super()._make_request(message, payload, protocol, _cls=_cls)
        ctx = self.make_context(req)
        ctx.log_request()
        req.set_context(ctx)
        return req
        
    def make_context(self, request):
        return Context(request)
    
    def make_handler(self, *, secure_proxy_ssl_header=None, **kwargs):
        """
            Copy of aiohttp.web_server.Server.make_handler
            with an option to define a cls for server
        """
        debug = kwargs.pop('debug', sentinel)
        if debug is not sentinel:
            warnings.warn(
                "`debug` parameter is deprecated. "
                "Use Application's debug mode instead", DeprecationWarning)
            if debug != self.debug:
                raise ValueError(
                    "The value of `debug` parameter conflicts with the debug "
                    "settings of the `Application` instance. The "
                    "application's debug mode setting should be used instead "
                    "as a single point to setup a debug mode. For more "
                    "information please check "
                    "http://aiohttp.readthedocs.io/en/stable/"
                    "web_reference.html#aiohttp.web.Application"
                )
        self.freeze()
        self._secure_proxy_ssl_header = secure_proxy_ssl_header
        return self.SERVER_CLS(self._handle, request_factory=self._make_request,
                               debug=self.debug, loop=self.loop, **kwargs)
    

class KtsContainerHttpApplication(KtsHttpApplication):
    def __init__(self, config, container=None, **kwargs):
        super().__init__(**kwargs)
        
        self.config = config
        self._container = weakref.ref(container) if container else None
    
    @property
    def container(self):
        return self._container() if self._container else None



from aiokts.util.arguments import Argument, ArgumentException, check_arguments


class ListArg(Argument):
    def __init__(self, required=True, default=[], allow_empty=False, max_count=None):
        validator = None
        validator_message = ''
        if not allow_empty and max_count is None:
            def validator(x):
                return len(x) > 0
            validator_message = 'must be non-empty list.'
        elif max_count is not None:
            def validator(x):
                return 0 < len(x) < max_count
            validator_message = 'must be non-empty list with max_count = {}.'.format(max_count)
        elif allow_empty and max_count is not None:
            def validator(x):
                return len(x) < max_count
            validator_message = 'must be list with max_count = {}.'.format(max_count)
        else:
            pass
        
        super().__init__(
            required=required,
            type=list,
            default=default,
            validator=validator,
            validator_message=validator_message,
            filter=None
        )


class StringListArg(Argument):
    def __init__(self, allowed_fields=None, required=True, default=[], allow_empty=False):
        super().__init__(
            required=required,
            type=list,
            default=default,
            validator=(lambda x: len(x) > 0) if not allow_empty else None,
            validator_message='must be non-empty list. Valid fields: "%s"' % (
                '", "'.join(allowed_fields)) if allowed_fields is not None else '',
            filter=(lambda x: list(filter(lambda y: y in allowed_fields, x))) if allowed_fields is not None else None
        )


class IntArg(Argument):
    def __init__(self, required=True, default=0):
        super().__init__(
            required=required,
            type=int,
            default=default
        )


class PositiveIntArg(Argument):
    def __init__(self, required=True, default=0, max_value=None):
        super().__init__(
            required=required,
            type=int,
            default=default,
            validator=(lambda x: x > 0) if max_value is None else (lambda x: 0 < x <= max_value),
            validator_message='must be {}'.format('> 0' if max_value is None else '> 0 and <= {}'.format(max_value))
        )


class NonNegativeIntArg(Argument):
    def __init__(self, required=True, default=0, max_value=None):
        super().__init__(
            required=required,
            type=int,
            default=default,
            validator=(lambda x: x >= 0) if max_value is None else (lambda x: 0 <= x <= max_value),
            validator_message='must be {}'.format('> 0' if max_value is None else '> 0 and <= {}'.format(max_value))
        )


class BoolArg(Argument):
    def __init__(self, required=True, default=False):
        super().__init__(
            required=required,
            type=bool,
            default=default,
            validator=None,
            validator_message=None
        )


class FlagsArg(Argument):
    def __init__(self, flags, required=True, default=None):
        if not default:
            default = {}
        super().__init__(
            required=required,
            type=dict,
            default=default,
            filter=flags.dict_filter
        )


class StringArg(Argument):
    def __init__(self, required=True, strip=True, allow_empty=False, empty_is_none=False, default=None):
        if strip:
            if empty_is_none:
                def filter(x):
                    x = x.strip()
                    return None if x == '' else x
            else:
                def filter(x):
                    return x.strip()
        else:
            if empty_is_none:
                def filter(x):
                    return None if x == '' else x
            else:
                filter = None
        super().__init__(
            required=required,
            type=str,
            filter=filter,
            validator=(lambda x: len(x) > 0) if not allow_empty else None,
            validator_message='must be non-empty string' if not allow_empty else None
        )


class ListOfArg(Argument):
    def __init__(self, required=True, argument_inst=None):
        validator = None
        filter = None
        if argument_inst is not None:
            if not isinstance(argument_inst, Argument):
                raise ArgumentException("ListOfArg's argument_type must be an instance of Argument class")
            
            if callable(argument_inst.validator):
                def validator(x):
                    for i, el in enumerate(x):
                        validation_res = argument_inst.validator(el)
                        if not validation_res:
                            self.validator_message = \
                                "Validation for element #{} failed: {}".format(i, argument_inst.validator_message)
                            return False
                        return True
            
            if callable(argument_inst.filter):
                def filter(x):
                    filtered = []
                    for el in x:
                        filtered.append(argument_inst.filter(el))
                    return filtered
        
        super().__init__(
            required=required,
            type=list,
            filter=filter,
            validator=validator,
            validator_message=None
        )


class DictArg(Argument):
    def __init__(self, required=True, required_fields=None, filter=None):
        if required_fields is not None:
            def validator(x):
                missing_fields = []
                for f in required_fields:
                    if x.get(f) is None:
                        missing_fields.append(f)
                if len(missing_fields) > 0:
                    self.validator_message = "argument's required fields {} are missing".format(missing_fields)
                    return False
                return True
        else:
            validator = None
        super().__init__(
            required=required,
            type=dict,
            filter=filter,
            validator=validator,
            validator_message=None
        )


class DictWithSchemaArg(Argument):
    """
        schema - Схема в виде
        {
            "field1": Argument(...),
            ...
        }
    """
    
    def __init__(self, schema, required=True, max_size=None):
        if schema is None:
            raise ArgumentException("Schema cannot be None in {}".format(str(self.__class__)))
        
        validator = None
        if max_size is not None:
            def validator(d):
                self.validator_message = 'Dict size must be less than {}'.format(max_size)
                return len(d) <= max_size
        
        def filter(d):
            return check_arguments(schema, d)
        
        super().__init__(
            required=required,
            type=dict,
            filter=filter,
            validator=validator,
            validator_message=None
        )


class ListWithSchemaArg(Argument):
    """
        schema - Схема в виде
        {
            "field1": Argument(...),
            ...
        }
    """
    
    def __init__(self, schema, required=True, max_size=None, empty_is_none=False, default=None):
        if schema is None:
            raise ArgumentException("Schema cannot be None in {}".format(str(self.__class__)))
        
        validator = None
        if max_size is not None:
            def validator(l):
                self.validator_message = 'List size must be less than {}'.format(max_size)
                return len(l) <= max_size
        
        def filter(l):
            if empty_is_none and len(l) == 0:
                return None
            filtered = []
            for el in l:
                el = check_arguments(schema, el)
                filtered.append(el)
            return filtered
        
        super().__init__(
            required=required,
            type=list,
            default=default,
            filter=filter,
            validator=validator,
            validator_message=None
        )


class ListOfDictsArg(Argument):
    def __init__(self, required=True, required_fields=None, empty_is_none=False):
        if required_fields is not None:
            def validator(x):
                for i, x_item in enumerate(x):
                    if not isinstance(x_item, dict):
                        self.validator_message = "all elements must be dicts (failed on element #{})".format(i)
                        return False
                    else:
                        missing_fields = []
                        for f in required_fields:
                            if x_item.get(f) is None:
                                missing_fields.append(f)
                        if len(missing_fields) > 0:
                            self.validator_message = \
                                "argument's required fields {} are missing in element #{}".format(missing_fields, i)
                            return False
                return True
        else:
            def validator(x):
                for i, x_item in enumerate(x):
                    if not isinstance(x_item, dict):
                        self.validator_message = "all elements must be dicts (failed on element #{})".format(i)
                        return False
                return True
        
        if empty_is_none:
            def filter(l):
                return None if len(l) == 0 else l
        else:
            filter = None
        
        super().__init__(
            required=required,
            type=list,
            filter=filter,
            validator=validator,
            validator_message=None
        )



class StoreSet(object):
    def __init__(self, config=None):
        self.stores = {}

        if config is None:
            config = {}

        for key, config in config.items():
            self.stores[key] = self.init_store(config)

    def __getattribute__(self, *args, **kwargs):
        store = super().__getattribute__('stores').get(args[0])
        if store is not None:
            return store
        return super().__getattribute__(*args, **kwargs)

    def init_store(self, config):
        type = config['type']
        utype = type[0].upper() + type[1:]
        name = '{0}Store'.format(utype)
        mod = __import__('aiokts.stores', fromlist=[name])

        return getattr(mod, name)(config)
